from typing import Iterable

from ..molecule import Bond


def get_bond_array(
    bonds: Iterable[Bond],
) -> str:

    bonds_js = ','.join(map(_bond_to_javascript, bonds))
    return f'[{bonds_js}]'


def _bond_to_javascript(bond: Bond) -> str:
    return (
        'md.bond'
        f'({bond.get_order()})'
        f'({bond.get_atom1_id()})'
        f'({bond.get_atom2_id()})'
    )



from __future__ import annotations

import os
from docutils import nodes
from sphinx.util.docutils import SphinxDirective
import pkgutil

from .molecule import (
    Molecule,
)
from .javascript import (
    get_atom_array,
    get_bond_array,
    get_scene_config,
    get_mesh_config,
)


class MolDocNode(nodes.Body, nodes.Element):
    def __init__(
        self,
        moldoc_name: str,
        molecule: Molecule,
    ) -> None:

        super().__init__()
        self._moldoc_name = moldoc_name
        self._molecule = molecule

    def get_molecule(self) -> Molecule:
        return self._molecule

    def get_moldoc_name(self) -> str:
        return self._moldoc_name


class MolDoc(SphinxDirective):
    has_content = True

    def run(self):
        content = '\n'.join(self.content)
        globals_ = {}
        exec(content, globals_)
        node = MolDocNode(
            moldoc_name=f'moldoc_{self.env.new_serialno("moldoc")}',
            molecule=globals_['moldoc_display_molecule'],
        )
        return [node]


def html_moldoc(self, node: MolDocNode):
    molecule = node.get_molecule()
    moldoc_node_id = node.get_moldoc_name()

    if not getattr(self, 'moldoc_scripts_added', False):
        self.body.append(
            '<script src="/_static/three.min.js"></script>'
            '<script src="/_static/molDraw.js"></script>'
            '<script>const md=molDraw;'
            'let atoms=[];'
            'let bonds=[];'
            'let maybeMolecule=undefined;'
            '</script>'
        )
        self.moldoc_scripts_added = True

    content = (
        f'atoms={get_atom_array(molecule.get_atoms())};'
        f'bonds={get_bond_array(molecule.get_bonds())};'
        'maybeMolecule=md.maybeMolecule(atoms)(bonds);'
        'if (md.isLeft(maybeMolecule))'
        '{'
        'console.log("There was an issue with your molecule.");'
        'console.log(md.fromLeft()(maybeMolecule));'
        '}'
        'else'
        '{'
        'const molecule=md.fromRight()(maybeMolecule);'
        'const scene=md.scene({'
        f'{get_scene_config(moldoc_node_id, molecule.get_config())}'
        '});'
        'const meshes=md.meshes({'
        f'{get_mesh_config(molecule)}'
        '})(molecule);'
        'md.drawMol(scene(meshes));'
        '}'
    )

    self.body.append(
        f'<div id="{moldoc_node_id}" style="height:25vh;"></div>'
        f'<script>{content}</script>'
    )
    raise nodes.SkipNode


def copy_asset_files(app, exc) -> None:
    asset_files = (
        'molDraw.js',
        'three.min.js',
    )
    static_dir = os.path.join(app.builder.outdir, '_static')
    # Build is HTML and succeeded.
    if app.builder.format == 'html' and exc is None:
        for path in asset_files:
            with open(os.path.join(static_dir, path), 'wb') as f:
                f.write(pkgutil.get_data(__package__, path))


def setup(app):
    app.connect('build-finished', copy_asset_files)
    app.add_directive('moldoc', MolDoc)
    app.add_node(
        node=MolDocNode,
        html=(html_moldoc, None),
    )
    return {
        'version': '0.0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



from typing import Optional, TypeVar

from .bool import bool_to_javascript
from ..molecule import (
    Material,
    MeshLambertMaterial,
    MeshNormalMaterial,
    MeshPhongMaterial,
    MeshPhysicalMaterial,
    MeshStandardMaterial,
    MeshToonMaterial,
)


def material_to_javascript(material: Material) -> str:
    # This is a type-safe switch
    return {  # type: ignore
        MeshLambertMaterial: _lambert_to_javascript,
        MeshNormalMaterial: _normal_to_javascript,
        MeshPhongMaterial: _phong_to_javascript,
        MeshPhysicalMaterial: _physical_to_javascript,
        MeshStandardMaterial: _standard_to_javascript,
        MeshToonMaterial: _toon_to_javascript,
    }[type(material)](material)


def _lambert_to_javascript(material: MeshLambertMaterial) -> str:
    return (
        'material:(color)=>'
        'new THREE.MeshLambertMaterial({color:color})'
    )


def _normal_to_javascript(material: MeshNormalMaterial) -> str:
    return 'material:(color)=>new THREE.MeshNormalMaterial()'


def _phong_to_javascript(
    material: MeshPhongMaterial,
) -> str:

    shininess = _property_to_javascript(
        property_name='shininess',
        property_value=material.get_shininess(),
    )
    reflectivity = _property_to_javascript(
        property_name='reflectivity',
        property_value=material.get_reflectivity(),
    )
    flat_shading = _property_to_javascript(
        property_name='flatShading',
        property_value=material.get_flat_shading(),
    )
    return (
        'material:'
        '(color)=>new THREE.MeshPhongMaterial({'
        'color:color'
        f'{"," if shininess else ""}{shininess}'
        f'{"," if reflectivity else ""}{reflectivity}'
        f'{"," if flat_shading else ""}{flat_shading}'
        '})'
    )


def _physical_to_javascript(
    material: MeshPhysicalMaterial,
) -> str:
    clearcoat = _property_to_javascript(
        property_name='clearcoat',
        property_value=material.get_clearcoat(),
    )
    clearcoat_roughness = _property_to_javascript(
        property_name='clearcoat_roughness',
        property_value=material.get_clearcoat_roughness(),
    )
    roughness = _property_to_javascript(
        property_name='roughness',
        property_value=material.get_roughness(),
    )
    metalness = _property_to_javascript(
        property_name='metalness',
        property_value=material.get_metalness(),
    )
    reflectivity = _property_to_javascript(
        property_name='reflectivity',
        property_value=material.get_reflectivity(),
    )
    flat_shading = _property_to_javascript(
        property_name='flat_shading',
        property_value=material.get_flat_shading(),
    )
    return (
        'material:'
        '(color)=>new THREE.MeshPhysicalMaterial({'
        'color:color'
        f'{"," if clearcoat else ""}{clearcoat}'
        f'{"," if clearcoat_roughness else ""}{clearcoat_roughness}'
        f'{"," if roughness else ""}{roughness}'
        f'{"," if metalness else ""}{metalness}'
        f'{"," if reflectivity else ""}{reflectivity}'
        f'{"," if flat_shading else ""}{flat_shading}'
        '})'
    )


def _standard_to_javascript(
    material: MeshStandardMaterial,
) -> str:
    metalness = _property_to_javascript(
        property_name='metalness',
        property_value=material.get_metalness(),
    )
    roughness = _property_to_javascript(
        property_name='roughness',
        property_value=material.get_roughness(),
    )
    flat_shading = _property_to_javascript(
        property_name='flatShading',
        property_value=material.get_flat_shading(),
    )
    return (
        'material:'
        '(color)=>new THREE.MeshStandardMaterial({'
        'color:color'
        f'{"," if metalness else ""}{metalness}'
        f'{"," if roughness else ""}{roughness}'
        f'{"," if flat_shading else ""}{flat_shading}'
        '})'
    )


def _toon_to_javascript(
    material: MeshToonMaterial,
) -> str:
    ...


T = TypeVar('T')


def _property_to_javascript(
    property_name: str,
    property_value: Optional[T],
) -> str:

    if property_value is None:
        return ''

    if isinstance(property_value, bool):
        return f'{property_name}:{bool_to_javascript(property_value)}'

    return f'{property_name}:{property_value}'



from .mesh_config import get_mesh_config
from .scene_config import get_scene_config
from .atoms import get_atom_array
from .bonds import get_bond_array


__all__ = [
    'get_mesh_config',
    'get_scene_config',
    'get_atom_array',
    'get_bond_array',
]
